
/***********************************************************************
 *                                                                     *
 *                          Table of Contents                          *
 *                                                                     *
 ***********************************************************************
  Introduction......................................................##A
  Application Object................................................##B 
  Application Init..................................................##C
    Activity Callback Installation.................................#C.1
    Worker Function................................................#C.2
  System Callbacks..................................................##D
    Application State..............................................#D.1
    System Object Handling.........................................#D.2
    Command Handler................................................#D.3
    Command Loop...................................................#D.4
  Basic Graphics....................................................##E
  Events............................................................##F
  Miscellaneous.....................................................##M
    Definitions....................................................#M.1
    Prototypes.....................................................#M.2
    Unimplemented..................................................#M.3
 ***********************************************************************/

/***********************************************************************
 *##A                                                                  *
 *                        Android Graphics I/O                         *
 *                                                                     *
 ***********************************************************************
  The android system is very weird in the way it handles events. First
  of all, when you make a native application, it's not really this native
  program running directly on the processor. A Java class is loaded first,
  which sets up the environment in which to talk to the rest of the system.
  This Java program then loads some methods from your native shared library
  and calls them. While it's doing these things, some threads are created
  to do various stuff, I don't really understand when and how many. 

  The system provides events thru several channels. There are two main
  methods used. One group of events are 'system events'. These are not
  related to user input etc. but are generated by the system in order to
  manage the application. For example, when the application is sent to
  background, a 'Pause' event is sent. Such system events are not reported
  thru a file descriptor but are processed directly by the Java wrapper
  around the native shared object. The application object exposed to the
  Java interface contains some function pointers, one for each type of
  system event. The Java part calls these function in *SOME* thread when
  the corresponding event occurs. This causes system events to be reported
  asynchronously. Your event loop in the main thread never sees these
  things, they get handled by another thread. In order to make them 
  synchronous, we report these events to the main thread via a pipe from
  the event handler functions.

  Another group of events are related to user input. These get reported
  thru a file descriptor, much like X11 events. The system has a struct
  called a 'looper' which can watch multiple file descriptors for events
  or 'ready to read' status. The looper is actually a simple wrapper 
  around the 'poll' system call. When you're ready to process user input,
  you add the input queue to the looper and then go into a loop, telling
  the looper to report activity on any file descriptor. 

  Fortunately, the looper can also watch regular file descriptors such
  as pipes or sockets. This makes it possible to synchronize system
  events with user events. 

  Here, we start with some boilerplate.
 ***********************************************************************/

#include <jni.h>
#include <math.h>

#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/resource.h>
#include <poll.h>
#include <pthread.h>
#include <sched.h>
#include <android/configuration.h>
#include <android/looper.h>
#include <android/native_activity.h>
#include <android/log.h>
#include "gfx.h"

#define LOGI(...) \
  ((void)__android_log_print(ANDROID_LOG_INFO, "threaded_app", __VA_ARGS__))
#define LOGE(...) \
  ((void)__android_log_print(ANDROID_LOG_ERROR, "threaded_app", __VA_ARGS__))

#ifndef NDEBUG
#  define LOGV(...) \
  ((void)__android_log_print(ANDROID_LOG_VERBOSE, "threaded_app", __VA_ARGS__))
#else
#  define LOGV(...)  ((void)0)
#endif


/***********************************************************************
 *#M.1                                                                 *
 *                             Definitions                             *
 *                                                                     *
 ***********************************************************************
  The PIPE_* constants indicate which channel is ready for reading.
  They are given as identifiers to the looper thingy.

    PIPE_CMD     indicates that some asynchronous event has happened and
                 the callback has written some event on the command pipe.
    PIPE_EVENT   indicates that user input is available.
    PIPE_SENSOR  will be used for the sensor channel. It's not used yet.
 ***********************************************************************/
#define  PIPE_CMD    1
#define  PIPE_EVENT  2
#define  PIPE_SENSOR 3

/***********************************************************************
  CMD_* constants identify which asynchronous callback has occured.
 ***********************************************************************/
  
enum
{
  CMD_APP_START,
  CMD_APP_STOP,
  CMD_APP_RESUME,
  CMD_APP_PAUSE,
  CMD_WINDOW_DESTROYED,
  CMD_WINDOW_CREATED,
  CMD_WINDOW_RESIZED,
  CMD_WINDOW_REDRAW,
  CMD_WINDOW_CONTENTRECT,
  CMD_SAVE_STATE,
  CMD_FOCUS_CHANGE,
  CMD_LOW_MEMORY,
  CMD_CONFIG_CHANGE,
  CMD_INPUTQ_DESTROYED,
  CMD_INPUTQ_CREATED,
  CMD_QUIT
};


/***********************************************************************
 *#M.2                                                                 *
 *                             Prototypes                              *
 *                                                                     *
 ***********************************************************************/
typedef struct req request_t;
typedef struct app app_t;

static void* worker_entry(void* arg);
static void onStop(ANativeActivity* activity);
static void onPause(ANativeActivity* activity);
static void onResume(ANativeActivity* activity);
static void onStart(ANativeActivity* activity);
static void onNativeWindowCreated
              (ANativeActivity* activity, ANativeWindow* window);
static void onNativeWindowDestroyed
              (ANativeActivity* activity, ANativeWindow* window);
static void onNativeWindowRedrawNeeded
              (ANativeActivity* activity, ANativeWindow* window);
static void onInputQueueCreated
              (ANativeActivity* activity, AInputQueue* queue);
static void onInputQueueDestroyed
              (ANativeActivity* activity, AInputQueue* queue);
static void* onSaveInstanceState
              (ANativeActivity* activity, size_t* outLen);
static void onWindowFocusChanged
              (ANativeActivity* activity, int focused);
static void onLowMemory(ANativeActivity* activity);
static void onConfigurationChanged(ANativeActivity* activity);

static void make_request(app_t *A, request_t *R);
static void complete_request(app_t *A, request_t *R);
static int get_request(app_t *A,request_t **R);
static int execute_commands(app_t *app);

static void graphics_window_destroyed(app_t *A, ANativeWindow *window);
static void graphics_window_created(app_t *A, ANativeWindow *window);

static int imax2(int a,int b);
static int imin2(int a,int b);
static int handle_input(app_t *A, gfx_event_t *ev);
static void clip_for_rect(int *start, int *amount, int max);

/***********************************************************************
 *##B                                                                  *
 *                         Application Object                          *
 *                                                                     *
 ***********************************************************************
  The application object is called an 'instance' in the android documents.
  It contains the necessary data to properly handle and serialize events.
  Window handle etc. are also contained in this object.
 ***********************************************************************/


struct app {
      // command pipe. This is used for reporting system events such
      // as Pause, Resume, etc.
  int cmdpipe_r, cmdpipe_w;

      // The worker thread is actually the 'main' thread for our
      // application. It initializes the input system and then calls
      // android_main. When that function is done, it rips up the
      // input/output objects and quits.
  pthread_t worker_thread;

      // These two are used for signalling various conditions. 
  pthread_mutex_t mutex;
  pthread_cond_t cond;

      // This is one of the conditions. When this is set to 1, 
      // the thread which fired off the worker thread can continue
      // in its Java world.
  int running;

  void *init_state; size_t init_state_size;

  ANativeActivity *activity;
  ALooper *looper;
  AConfiguration *config;
  AInputQueue *inputq;
  ANativeWindow *window;

  request_t *quit_request;
  int last_pointer_id;
};

static app_t *my_app;


/***********************************************************************
 *##C                                                                  *
 *                          Application Init                           *
 *                                                                     *
 ***********************************************************************
  During the initialization, we need to do the following:
   - create a pipe for the asynchronous messages coming in
   - create a mutex and a condition variable for protecting shared memory
   - create a worker thread

  The onCreate function is the entry point from the Java world. We 
  initialize the 'activity' object with the system event handler functions.
 ***********************************************************************/

void ANativeActivity_onCreate
      (ANativeActivity* activity, void* state, size_t state_size)
{
  int cmdpipe[2];
  pthread_attr_t attr; 
  app_t *app;

  LOGV("Creating: %p\n", activity);

  // #C.1: Activity callback installation
  // activity->callbacks->onDestroy = onDestroy;
  activity->callbacks->onStart = onStart;
  activity->callbacks->onResume = onResume;
  activity->callbacks->onPause = onPause;
  activity->callbacks->onStop = onStop;
  activity->callbacks->onNativeWindowCreated = onNativeWindowCreated;
  activity->callbacks->onNativeWindowDestroyed = onNativeWindowDestroyed;
  activity->callbacks->onInputQueueCreated = onInputQueueCreated;
  activity->callbacks->onInputQueueDestroyed = onInputQueueDestroyed;
  activity->callbacks->onSaveInstanceState = onSaveInstanceState;
  activity->callbacks->onWindowFocusChanged = onWindowFocusChanged;
  activity->callbacks->onLowMemory = onLowMemory;
  activity->callbacks->onConfigurationChanged = onConfigurationChanged;

  ANativeActivity_setWindowFormat(activity, WINDOW_FORMAT_RGBX_8888);

  activity->instance= my_app= app= calloc(1,sizeof(*app));
  app->activity= activity;
  app->last_pointer_id= -1;

  if (state && state_size)
  {
    app->init_state= malloc(state_size);
    memcpy(app->init_state, state, state_size);
    app->init_state_size= state_size;
  }

  pthread_mutex_init(&app->mutex, NULL);
  pthread_cond_init(&app->cond,NULL);

  if (pipe(cmdpipe))
  {
    LOGE("Couldn't create pipe: %s\n",strerror(errno));
    return ;
  }
  app->cmdpipe_r= cmdpipe[0];
  app->cmdpipe_w= cmdpipe[1];

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  pthread_create(&app->worker_thread, &attr, worker_entry, app);

  
  pthread_mutex_lock(&app->mutex); 
  while(!app->running) pthread_cond_wait(&app->cond, &app->mutex); 
  pthread_mutex_unlock(&app->mutex);
}

  // #C.2 Worker Function
static void* worker_entry(void* arg)
{
  app_t *app;

  app= arg;

  app->config= AConfiguration_new();
  AConfiguration_fromAssetManager(app->config, app->activity->assetManager);

  app->looper= ALooper_prepare(ALOOPER_PREPARE_ALLOW_NON_CALLBACKS);
  ALooper_addFd(app->looper, app->cmdpipe_r,
                PIPE_CMD, ALOOPER_EVENT_INPUT, NULL, app);
  
  pthread_mutex_lock(&app->mutex); 
  app->running=1; 
  pthread_cond_broadcast(&app->cond);
  pthread_mutex_unlock(&app->mutex);

  while(!app->inputq || !app->window)
    if (execute_commands(app)) goto done;

  android_main(app->init_state, app->init_state_size);

  while(!app->quit_request)
     if (execute_commands(app))
       break;

done:
  //free_saved_state(A);
  if (app->inputq) 
      { AInputQueue_detachLooper(app->inputq); app->inputq= NULL; }
  AConfiguration_delete(app->config);
  app->config= NULL;

  if (app->quit_request)
     complete_request(app, app->quit_request);
  return NULL;
}


/***********************************************************************
 *##D                                                                  *
 *                          System Callbacks                           *
 *                                                                     *
 ***********************************************************************
  System callbacks are things like onResume, onWindowCreated etc. These
  are supposed to be handled in a separate thread but that makes things
  complicated. So, I'm going to serialize them into a pipe and let the
  main thread do everything. The google guys made sure that these events
  are acknowledged by the main thread before returning from the callback.
  So, I'm going to do the same thing and include an 'acknowledgement'
  field in each request. This might be important for adding event queues
  to loopers etc. 

  Requests are represented by the following struct. Only a pointer to 
  this struct is passed around since everything is in the same address
  space.
 ***********************************************************************/

struct req {
  int type;
  int complete;
  union {
    ANativeWindow *window;
    AInputQueue   *inputq;
    struct {
       void* ptr;
       size_t size;
    }             buffer; 
    int           intv;
  } d;
};

/***********************************************************************
  The following functions send, receive and acknowledge requests.
 ***********************************************************************/

static void make_request(app_t *A, request_t *R)
{
   R->complete= 0;
   if (write(A->cmdpipe_w, &R, sizeof(R))!=sizeof(R))
   {
     LOGE("Couldn't write to command pipe: %s\n", strerror(errno));
     return ;
   }
   
   pthread_mutex_lock(&A->mutex); 
   while(!R->complete) pthread_cond_wait(&A->cond, &A->mutex); 
   pthread_mutex_unlock(&A->mutex); 
}

static int get_request(app_t *A,request_t **R)
{
   request_t *T;
   if (read(A->cmdpipe_r, &T, sizeof(T))!=sizeof(T))
   {
     LOGE("Couldn't read from command pipe: %s\n", strerror(errno));
     return 1;
   }
   *R= T;
   return 0;
}

static void complete_request(app_t *A, request_t *R)
{
   pthread_mutex_lock(&A->mutex); 
   R->complete= 1;
   pthread_cond_broadcast(&A->cond);
   pthread_mutex_unlock(&A->mutex); 
}


/***********************************************************************
 * #D.1                                                                *
 *                          Application State                          *
 *                                                                     *
 ***********************************************************************/

static void simple_command(ANativeActivity *ac, int cmd)
{
   request_t R;
   R.type= cmd;
   make_request(ac->instance, &R);
}

static void onStart(ANativeActivity* activity)
{
   LOGV("Start: %p\n", activity);
   simple_command(activity, CMD_APP_START);
}

static void onResume(ANativeActivity* activity)
{
   LOGV("Resume: %p\n", activity);
   simple_command(activity, CMD_APP_RESUME);
}

static void onPause(ANativeActivity* activity)
{
   LOGV("Pause: %p\n", activity);
   simple_command(activity, CMD_APP_PAUSE);
}

static void onStop(ANativeActivity* activity)
{
   LOGV("Stop: %p\n", activity);
   simple_command(activity, CMD_APP_STOP);
}

/***********************************************************************
 *#D.2                                                                 *
 *                       System Object Handling                        *
 *                                                                     *
 ***********************************************************************
  The system notifies the availability of various objects such as the
  native window, input queue etc. thru callbacks. It's quite stupid, but
  what can you do? The destruction of these objects are also reported
  in the same way.
 
  It looks like the default glue code doesn't handle all of the callbacks.
  These are: 

  void (*onNativeWindowResized)
        (ANativeActivity* activity, ANativeWindow* window);
  void (*onContentRectChanged)
        (ANativeActivity* activity, const ARect* rect);

  WindowResized and ContentRectChanged are called when the device is
  rotated, as far as I understand. WindowRedrawNeeded is called when
  the application is brought back to foreground. I will do these after 
  I have the graphics system.
 ***********************************************************************/

static void onNativeWindowCreated
   (ANativeActivity* activity, ANativeWindow* window)
{
  request_t R;
  LOGV("NativeWindowCreated: %p -- %p\n", activity, window);
  R.type= CMD_WINDOW_CREATED;
  R.d.window= window;
  make_request(activity->instance, &R);
}

static void onNativeWindowDestroyed
   (ANativeActivity* activity, ANativeWindow* window)
{
  request_t R;
  LOGV("NativeWindowDestroyed: %p -- %p\n", activity, window);
  R.type= CMD_WINDOW_DESTROYED;
  R.d.window= window;
  make_request(activity->instance, &R);
}

static void onNativeWindowRedrawNeeded
              (ANativeActivity* activity, ANativeWindow* window)
{
  request_t R;
  LOGV("NativeWindowRedraw: %p -- %p\n", activity, window);
  R.type= CMD_WINDOW_REDRAW;
  R.d.window= window;
  make_request(activity->instance, &R);
}

static void onInputQueueCreated
   (ANativeActivity* activity, AInputQueue* queue)
{
  request_t R;
  LOGV("InputQueueCreated: %p -- %p\n", activity, queue);
  R.type= CMD_INPUTQ_CREATED;
  R.d.inputq= queue;
  make_request(activity->instance, &R);
}

static void onInputQueueDestroyed
   (ANativeActivity* activity, AInputQueue* queue)
{
  request_t R;
  LOGV("InputQueueDestroyed: %p -- %p\n", activity, queue);
  R.type= CMD_INPUTQ_DESTROYED;
  R.d.inputq= queue;
  make_request(activity->instance, &R);
}

static void* onSaveInstanceState
   (ANativeActivity* activity, size_t* outLen)
{
  request_t R;
  LOGV("SaveInstanceState: %p\n", activity);
  R.type= CMD_SAVE_STATE;
  R.d.buffer.ptr= NULL;
  R.d.buffer.size= 0;
  make_request(activity->instance, &R);
  if (R.d.buffer.ptr)
    *outLen= R.d.buffer.size;
  return R.d.buffer.ptr;
}

static void onWindowFocusChanged
   (ANativeActivity* activity, int focused)
{
  request_t R;
  LOGV("WindowFocusChanged: %p -- %d\n", activity, focused);
  R.type= CMD_FOCUS_CHANGE;
  R.d.intv= focused;
  make_request(activity->instance, &R);
}

static void onLowMemory(ANativeActivity* activity)
{
  request_t R;
  LOGV("LowMemory: %p\n", activity);
  R.type= CMD_LOW_MEMORY;
  make_request(activity->instance, &R);
}

static void onConfigurationChanged(ANativeActivity* activity)
{
  request_t R;
  LOGV("ConfigurationChanged: %p\n", activity);
  R.type= CMD_CONFIG_CHANGE;
  make_request(activity->instance, &R);
}


/***********************************************************************
 *#D.3                                                                 *
 *                           Command Handler                           *
 *                                                                     *
 ***********************************************************************
  The following function handles the various commands generated by the
  callbacks. Some things are still missing, such as reporting the 
  application state etc. I will do these later on.

  The function returns 1 if the received command causes us to quit the
  application. This happens only if we get a destroy callback. This 
  callback isn't implemented yet.
 ***********************************************************************/

static int handle_command(app_t *A)
{
  int cmd; void *ptr;
  request_t *R;
  if (get_request(A, &R)) return 0;
  switch(R->type)
  {
  case CMD_APP_STOP:
  case CMD_APP_START:
  case CMD_APP_PAUSE:
  case CMD_APP_RESUME:
     // FIXME: inform the client, maybe??
     complete_request(A,R);
     break;
  case CMD_WINDOW_DESTROYED:
     graphics_window_destroyed(A, R->d.window);
     complete_request(A,R);
     break;
  case CMD_WINDOW_CREATED:
     graphics_window_created(A, R->d.window);
     complete_request(A,R);
     gfx_update_all();
     break;
  case CMD_WINDOW_REDRAW:
     gfx_update_all();
     break;
  case CMD_SAVE_STATE:
  case CMD_FOCUS_CHANGE:
  case CMD_LOW_MEMORY:
     complete_request(A,R);
     break;
  case CMD_CONFIG_CHANGE:
     LOGV("APP_CMD_CONFIG_CHANGED\n");
     AConfiguration_fromAssetManager(A->config, A->activity->assetManager);
     complete_request(A,R);
     break;
  case CMD_INPUTQ_DESTROYED:
     if (A->inputq!=R->d.inputq)
     {
       complete_request(A,R);
       break;
     }
     AInputQueue_detachLooper(A->inputq);
     A->inputq= NULL;
     complete_request(A,R);
     break;
  case CMD_INPUTQ_CREATED:
     if (A->inputq!=NULL)
        AInputQueue_detachLooper(A->inputq);
     A->inputq= R->d.inputq;
     AInputQueue_attachLooper(A->inputq, A->looper, PIPE_EVENT, NULL,A);
     complete_request(A,R);
     break;
  case CMD_QUIT:
     // FIXME: when do we get this?
     LOGV("APP_CMD_QUIT\n");
     A->quit_request= R;
     return 1;
  }
  return 0;
}

/***********************************************************************
 *#D.4                                                                 *
 *                            Command Loop                             *
 *                                                                     *
 ***********************************************************************
  We need a function which simply waits for commands and ignores events.
  This is necessary when we're initalizing or destroying the application
  object. The function returns 1 if a request to quit the application was
  received. Otherwise, it returns 0.
 ***********************************************************************/

static int execute_commands(app_t *app)
{
   int ident;
   int dFD, dEV;
   void *dPTR;

again:
   ident= ALooper_pollAll(-1,  &dFD, &dEV,  &dPTR);

   switch(ident)
   {
     // somebody might have woke us up because the looper
     // has changed and we need to poll again.
   case ALOOPER_POLL_WAKE:
      goto again;
   case PIPE_CMD:
      if (handle_command(app)) return 1;
      return 0;
   case PIPE_EVENT:
     handle_input(app, NULL);
     goto again;
   }

   // other things are not possible. maybe I can add some
   // sort of a time-out later.
   // so, unhandled stuff corresponds to an error, which
   // makes the application quit. I think it's OK.
   return 1;
}

/***********************************************************************
 *#M.3                                                                 *
 *                            Unimplemented                            *
 *                                                                     *
 ***********************************************************************
  The following stuff isn't used yet. I will incorporate them later,
  when I have most of the stuff working properly.

  The destruction function is supposedly executed by the 'main' thread,
  which is the thread that called onCreate initially.
 ***********************************************************************/

#if 0
static void onDestroy(ANativeActivity* activity)
{
  request_t R;
  app_t *A;

  LOGV("Destroy: %p\n", activity);

  A= activity->instance;
  R.type= CMD_QUIT;
  make_request(A, &R);

  close(A->cmdpipe_r);
  close(A->cmdpipe_w);
  pthread_cond_destroy(&A->cond);
  pthread_mutex_destroy(&A->mutex);
  free(A);
}

static void free_saved_state(struct android_app* android_app) {
    pthread_mutex_lock(&android_app->mutex);
    if (android_app->savedState != NULL) {
        free(android_app->savedState);
        android_app->savedState = NULL;
        android_app->savedStateSize = 0;
    }
    pthread_mutex_unlock(&android_app->mutex);
}

#endif

/***********************************************************************
 *##E                                                                  *
 *                           Basic Graphics                            *
 *                                                                     *
 ***********************************************************************
  In contrast to the input system, the output system is quite simple.
  When you want to put something on the screen, you
   - lock the corresponding rectangle
   - fill the returned buffer
   - unlock and post the buffer
  Since android doesn't keep the contents of a window, I need to do it
  myself, just like I did it for the X11 version, using an X11 image
  and pixmap combo. I will store the image as an array of uint32_t pixels,
  but I'm planning to update the screen using other formats as well.
 
  So basicly, the image will be stored in RGBX_8888 format but if the
  unlock stuff returns a different format, I will convert the pixels
  to that format when I'm writing to the pixel array of the buffer.
 ***********************************************************************/

static uint32_t *Gpixels, x_window_width, x_window_height, **x_image_data;

#define x_red_mask    0x0000ff
#define x_green_mask  0x00ff00
#define x_blue_mask   0xff0000
#define x_red_shift   0
#define x_green_shift 8
#define x_blue_shift  16

static void graphics_window_destroyed(app_t *A, ANativeWindow *window)
{
  if (A->window!=window) return ;
  A->window= NULL;
}

static void graphics_window_created(app_t *A, ANativeWindow *window)
{
   int W,H;
   A->window= window;
   W= ANativeWindow_getWidth(A->window);
   H= ANativeWindow_getHeight(A->window);
   if (W<=0) W= 1;
   if (H<=0) H= 1;

   if (W!=x_window_width || H!=x_window_height || !Gpixels)
   {
     int y;
     if (Gpixels) free(Gpixels);
     Gpixels= malloc(sizeof(uint32_t)*W*H);
     if (x_image_data) free(x_image_data);
     x_image_data= malloc(sizeof(x_image_data[0])*H);
     for(y=0;y<H;y++)
       x_image_data[y]= y*W + Gpixels;
     x_window_width= W;
     x_window_height= H;
   }
}


void gfx_update(int x,int y,int w,int h)
{
   ARect rect;
   ANativeWindow_Buffer buffer;
   int R;
   uint32_t *spixels;
   uint16_t *wpixels;
   int eX, eY;

   if (!Gpixels) return ; // display not initialized yet

   eX= x+w; eY= y+h;

   if (x>=x_window_width || eX<=0 || y>=x_window_height || eY<=0) return ;

   eX= imin2(eX, x_window_width); eY= imin2(eY, x_window_height);
   x= imax2(x, 0); y= imax2(y,0);

   rect.left= x; rect.top= y;
   rect.right= eX; rect.bottom= eY;

   R= ANativeWindow_lock(my_app->window, &buffer, &rect);
   if (R<0)
   {
      LOGV("Unable to lock buffer: %d\n", R);
      return ;
   }

   // LOGV("GFX::Buffer width=%d height=%d stride=%d format=%d\n", buffer.width, buffer.height,  buffer.stride, buffer.format);

   // LOGV("GFX::Rect left=%d top=%d right=%d bottom=%d\n", rect.left,
           // rect.top, rect.right, rect.bottom);
   
  switch(buffer.format)
  {
   case WINDOW_FORMAT_RGBX_8888:
     spixels= buffer.bits;
     for(y=rect.top;y<rect.bottom;y++)
       for(x=rect.left;x<rect.right;x++)
         spixels[y*buffer.stride+x]= Gpixels[x+ y*x_window_width];
     break;
   case WINDOW_FORMAT_RGBA_8888:
     spixels= buffer.bits;
     for(y=rect.top;y<rect.bottom;y++)
       for(x=rect.left;x<rect.right;x++)
         spixels[y*buffer.stride+x]= Gpixels[x+ y*x_window_width] | 0xff000000;
     break;
   case WINDOW_FORMAT_RGB_565:
     wpixels= buffer.bits;
     for(y=rect.top;y<rect.bottom;y++)
       for(x=rect.left;x<rect.right;x++)
       {
         uint32_t P,R,G,B;
         P= x_image_data[y][x];
         R= (P>>16)&0xff;  R>>= 3;
         G= (P>>8) &0xff;  G>>= 2;
         B= P&0xff;        B>>= 3;
         wpixels[y*buffer.stride+x]= (R<<11) | (G<<5) | B;
       }
      break;
    default:
      LOGV("unknown buffer format %d",buffer.format);
      break;
  }
  ANativeWindow_unlockAndPost(my_app->window);
}

void gfx_update_all()
{
  gfx_update(0,0,x_window_width,x_window_height);
}

/***********************************************************************
 *##M                                                                  *
 *                            Miscellaneous                            *
 *                                                                     *
 ***********************************************************************
  This section defines several utility functions. I wonder if these get
  inlined when compiled with optimization. It looks like they do, for
  gcc -O3.
 ***********************************************************************/

static int imax2(int a,int b) { return a>b?a:b; }
static int imin2(int a,int b) { return a<b?a:b; }

static void clip_for_rect(int *start, int *amount, int max)
{
  int end;
  if (*amount<=0) return ;
  end= *start + *amount;
  if (end>max) end= max;
  if (*start<0) *start= 0;
  *amount= end - *start;
}

/***********************************************************************
 *##F                                                                  *
 *                               Events                                *
 *                                                                     *
 ***********************************************************************
  The following event code is just a place holder for now. I'll look
  further into it later.
 ***********************************************************************/

static int handle_input(app_t *A, gfx_event_t *ev)
{
  AInputEvent* event;
  int type;
  uint32_t action,index, i, npointers;
  gfx_event_t dummy;
  event = NULL;
  if (!ev) ev= &dummy;
again:
     // Is this an error? should I log it or somehow
     // return an indication? FIXME

  if (AInputQueue_getEvent(A->inputq, &event)<0) return 1;

  type= AInputEvent_getType(event);
  // LOGV("New input event: type=%d\n", type);

  if (AInputQueue_preDispatchEvent(A->inputq, event)) goto again;

  switch(type)
  {
  case AINPUT_EVENT_TYPE_KEY: goto next_event;
  case AINPUT_EVENT_TYPE_MOTION: goto touch_event;
  }

next_event:
  AInputQueue_finishEvent(A->inputq, event, 1);
  goto again;

touch_event:
  action= AMotionEvent_getAction(event);
  index= (action & AMOTION_EVENT_ACTION_POINTER_INDEX_MASK)
              >>AMOTION_EVENT_ACTION_POINTER_INDEX_SHIFT;
  action= (action & AMOTION_EVENT_ACTION_MASK);

  switch(action)
  {
  case AMOTION_EVENT_ACTION_UP:
     ev->type= GFXEV_BUTTON_RELEASE;
     ev->button= 1;
     ev->x= AMotionEvent_getX(event,index);
     ev->y= AMotionEvent_getY(event,index);
     // LOGV("Mouse up: %d %d\n", ev->mouse.x, ev->mouse.y);
     A->last_pointer_id= -1;
     break;
  case AMOTION_EVENT_ACTION_DOWN:
     ev->type= GFXEV_BUTTON_PRESS;
     ev->x= AMotionEvent_getX(event,index);
     ev->y= AMotionEvent_getY(event,index);
     A->last_pointer_id= AMotionEvent_getPointerId(event, index);    
     // LOGV("Mouse down: %d %d\n", ev->mouse.x, ev->mouse.y);
     break;
  case AMOTION_EVENT_ACTION_MOVE:
     npointers= AMotionEvent_getPointerCount(event);
     for(i=0;i<npointers;i++)
        if (AMotionEvent_getPointerId(event, i)==A->last_pointer_id)
        { 
          ev->type= GFXEV_POINTER_MOVE;
          ev->x= AMotionEvent_getX(event,i);
          ev->y= AMotionEvent_getY(event,i);
          // LOGV("Mouse move: %d %d [%d %d %d]\n", ev->mouse.x, ev->mouse.y,
                             // npointers, i, A->last_pointer_id);
          goto done;
        }
  default:
     goto next_event;
  }

done:
  AInputQueue_finishEvent(A->inputq, event, 1);
  return 0;
}

int gfx_event(unsigned int *msecs, gfx_event_t *E)
{
   int ident;
   int dFD, dEV;
   app_t *app;
again:
   ident= ALooper_pollAll(-1,  &dFD, &dEV, (void**) &app);

   switch(ident)
   {
     // somebody might have woke us up because the looper
     // has changed and we need to poll again.
   case ALOOPER_POLL_WAKE:
      goto again;
   case PIPE_CMD:
      if (handle_command(app)) // quit was requested.
      {
        E->type= GFXEV_QUIT;
        return 1;
      }
      goto again;
   case PIPE_EVENT:
     if (handle_input(app, E) || !app->window)
       goto again;
     return 1;
   }

   // other things are not possible. maybe I can add some
   // sort of a time-out later.
   return 0;
}
